% FLENG compiler - inlining

collect_inlining_candidates([]).
collect_inlining_candidates([pdef(N/A, _, _, [C])|Gs]) :-
    \+module_init(N),
    C = t((H :- B), _, _),
    callable(B),
    H =.. [_|Args],
    functor(B, BN, BA),
    inlineable_operation(BN/BA),
    unique_vars_only(Args, [], Vars),
    assertz(inline_candidate(N/A, Vars, B)),
    collect_inlining_candidates(Gs).
collect_inlining_candidates([_|Gs]) :-
    collect_inlining_candidates(Gs).

inlineable_operation(N/A) :- pdef(N/A, _).
inlineable_operation(NA) :- primitive(NA, _, _, _).
inlineable_operation(NA) :-
    memberchk(NA, [true/0, '$asm'/_, foreign_call/1, compute/3, compute/4, ('=')/2,
        unify/3, call/1, call/2, call/3, '$call'/2, '$call_with_remote_task'/3,
        '$rcall'/2, '$rcall'/4, current_module/1]).

unique_vars_only([], _, []).
unique_vars_only(['$VAR'(I)|Args], Seen, [I|Tail]) :-
    \+memberchk(I, Seen),
    unique_vars_only(Args, [I|Seen], Tail).

reduce_inline_calls(X, X2, Ss1, Ss) :-
    extract_variables(X, [], Vars),
    inline_calls(X, X1, Vars, Ss1, _, Ss2),
    (X \== X1 -> 
        debug_compiler(inline, X, X1),
        reduce_inline_calls(X1, X2, Ss2, Ss)
    ;   X2 = X1, Ss = Ss2),
    !.

inline_calls((X, Y), (X2, Y2), Env1, Ss1, Env, Ss) :-
    inline_calls(X, X2, Env1, Ss1, E2, Ss2),
    inline_calls(Y, Y2, E2, Ss2, Env, Ss).
inline_calls(X, X2, Env1, Ss1, Env, Ss) :-
    callable(X),
    functor(X, N, A),
    X =.. [_|Args],
    inline_candidate(N/A, Vars, Body),
    map_argument_variables(Vars, Args, VarMap),
    substitute_variables(Body, VarMap, Env1, [], X2, Env, Ss2),
    append(Ss1, Ss2, Ss).
inline_calls(X, X, Env, Ss, Env, Ss).

extract_variables(X, Vs, Vs) :- atomic(X), !.
extract_variables('$VAR'(V), Vs, [V|Vs]) :- !.
extract_variables([X|Y], Vs1, Vs) :-
    !,
    extract_variables(X, Vs1, Vs2),
    extract_variables(Y, Vs2, Vs).
extract_variables(X, Vs1, Vs) :-
    callable(X),
    X =.. [_|Args],
    extract_variables(Args, Vs1, Vs).

map_argument_variables([], _, []).
map_argument_variables([V|Vars], [A|Args], [V-A|VarMap]) :-
    map_argument_variables(Vars, Args, VarMap).

substitute_variables(X, _, E, Ss, X, E, Ss) :- atomic(X).
substitute_variables('$VAR'(V), Map, E, Ss, Arg, E, Ss) :-
    memberchk(V-Arg, Map).
substitute_variables('$VAR'(_), _, E1, Ss, '$VAR'(V2), [V2|E1], [V2|Ss]) :-
    new_variable(0, E1, V2).
substitute_variables([X|Y], Map, E1, Ss1, [X2|Y2], E, Ss) :-
    substitute_variables(X, Map, E1, Ss1, X2, E2, Ss2),
    substitute_variables(Y, Map, E2, Ss2, Y2, E, Ss).
substitute_variables(X, Map, E1, Ss1, X2, E, Ss) :-
    callable(X),
    X =.. [H|Args],
    substitute_variables(Args, Map, E1, Ss1, Args2, E, Ss),
    X2 =.. [H|Args2].

new_variable(V, E, V) :- \+memberchk(V, E).
new_variable(V, E, NV) :-
    V2 is V + 1,
    new_variable(V2, E, NV).
