# FLENG - asm part of aarch64 runtime

    .section .text

    .global fl_check_atom
    // x0=val
fl_check_atom:
    and x8, x0, #3
    cmp x8, #2
    bne l1
    br x30
l1:
    mov x1, x0
    mov x0, x19
    mov x2, #1          // FL_NOT_AN_ATOM
    b fl_rt_error

    .global fl_check_cell
    // x0=val
fl_check_cell:
    tst x0, #3
    bne l2
    br x30
l2:
    mov x1, x0
    mov x0, x19
    mov x2, #2          // FL_NOT_A_CELL
    b fl_rt_error

    .global fl_check_integer
    // x0=val
fl_check_integer:
    tst x0, #1         // FL_INT_BIT
    beq l3
    br x30
l3:
    mov x1, x0
    mov x0, x19
    mov x2, #3          // FL_NOT_AN_INT
    b fl_rt_error

    .global fl_check_list
    // x0=val
fl_check_list:
    tst x0, #3
    bne l4
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x01000000 // FL_LIST_TAG
    cmp w8, w9
    bne l4
    br x30
l4:
    mov x1, x0
    mov x0, x19
    mov x2, #4          // FL_NOT_A_LIST
    b fl_rt_error

    .global fl_check_module
    // x0=val
fl_check_module:
    tst x0, #3
    bne l4b
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x0a000000 // FL_MODULE_TAG
    cmp w8, w9
    bne l4b
    br x30
l4b:
    mov x1, x0
    mov x0, x19
    mov x2, #23          // FL_NOT_A_MDOULE
    b fl_rt_error

    .global fl_check_tuple
    // x0=val
fl_check_tuple:
    tst x0, #3
    bne l5
    ldr w8, [x0]
    mov w9, #0x80000000  // FL_TUPLE_BIT
    tst w8, w9
    bne l5
    br x30
l5:
    mov x1, x0
    mov x0, x19
    mov x2, #5          // FL_NOT_A_TUPLE
    b fl_rt_error

    .global fl_check_var
    // x0=val
fl_check_var:
    tst x0, #3
    bne l6
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x04000000 // FL_VAR_TAG
    cmp w8, w9
    beq l6
    mov w9, #0x07000000 // FL_REF_TAG
    cmp w8, w9
    beq l6
    mov x1, x0
    mov x0, x19
    mov x2, #6          // FL_NOT_A_VAR
    b fl_rt_error
l6:
    br x30

    .global fl_check_number
    // x0=val
fl_check_number:
    tst x0, #1
    beq l7
    br x30
l7:
    tst x0, #2
    bne l8
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x02000000 // FL_FLOAT_TAG
    cmp w8, w9
    bne l8
    br x30
l8:
    mov x1, x0
    mov x0, x19
    mov x2, #11          // FL_NOT_A_NUMBER
    b fl_rt_error

    // x1=val -> d0
float_arg_1:
    tst x1, #1
    beq l9
    asr x1, x1, #1
    scvtf d0, x1
    br x30
l9:
    ldr d0, [x1, #8]
    br x30

    // x2=val -> d1
float_arg_2:
    tst x2, #1
    beq l10
    asr x2, x2, #1
    scvtf d1, x2
    br x30
l10:
    ldr d1, [x2, #8]
    br x30

    .global fl_add
    // x1=num, x2=num -> x0
fl_add:
    tst x1, #1
    beq l11
    tst x2, #1
    beq l11
    sub x0, x1, #1
    add x0, x0, x2
    br x30
l11:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    fadd d0, d0, d1
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_sub
    // x1=num, x2=num -> x0
fl_sub:
    tst x1, #1
    beq l12
    tst x2, #1
    beq l12
    sub x1, x1, x2
    add x0, x1, #1
    br x30
l12:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    fsub d0, d0, d1
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_mul
    // x1=num, x2=num -> x0
fl_mul:
    tst x1, #1
    beq l13
    tst x2, #1
    beq l13
    asr x1, x1, #1
    asr x2, x2, #1
    mul x0, x1, x2
    lsl x0, x0, #1
    orr x0, x0, #1
    br x30
l13:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    fmul d0, d0, d1
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_div
    // x1=num, x2=num -> x0
fl_div:
    cmp x2, #1
    bne l14
    mov x0, x19
    mov x2, #12         // FL_DIV_BY_ZERO
    b fl_rt_error
l14:
    tst x1, #1
    beq l15
    tst x2, #1
    beq l15
    asr x1, x1, #1
    asr x2, x2, #1
    sdiv x0, x1, x2
    lsl x0, x0, #1
    orr x0, x0, #1
    br x30
l15:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    fdiv d0, d0, d1
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_mod
    // x1=num, x2=num -> x0
fl_mod:
    cmp x2, #1
    bne l14b
    mov x0, x19
    mov x2, #12         // FL_DIV_BY_ZERO
    b fl_rt_error
l14b:
    asr x1, x1, #1
    asr x2, x2, #1
    sdiv x0, x1, x2
    msub x0, x2, x0, x1
    lsl x0, x0, #1
    orr x0, x0, #1
    br x30

    .global fl_pow
    // x1=num, x2=num -> x0
fl_pow:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    bl pow
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_sqrt
    // x1=num -> x0
fl_sqrt:
    str x30, [x28, #-8]!
    bl float_arg_1
    fsqrt d0, d0
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_sin
    // x1=num -> x0
fl_sin:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl sin
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_cos
    // x1=num -> x0
fl_cos:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl cos
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_tan
    // x1=num -> x0
fl_tan:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl tan
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_atan
    // x1=num -> x0
fl_atan:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl atan
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_loge
    // x1=num -> x0
fl_loge:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl log
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_exp
    // x1=num -> x0
fl_exp:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl exp
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_truncate
    // x1=num -> x0
fl_truncate:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl trunc
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_round
    // x1=num -> x0
fl_round:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl round
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_floor
    // x1=num -> x0
fl_floor:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl floor
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_ceiling
    // x1=num -> x0
fl_ceiling:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl ceil
    ldr x30, [x28], #8
    b fl_mkfloat

    .global fl_shl
    // x1=num, x2=num -> x0
fl_shl:
    lsr x2, x2, #1
    bic x1, x1, #1
    lsl x0, x1, x2
    orr x0, x0, #1
    br x30

    .global fl_shr
    // x1=num, x2=num -> x0
fl_shr:
    lsr x2, x2, #1
    asr x0, x1, x2
    orr x0, x0, #1
    br x30

    .global fl_int
    // x1=num -> x0
fl_int:
    tst x1, #1
    beq l16
    mov x0, x1
    br x30
l16:
    ldr d0, [x1, #8]
    fcvtzs x0, d0
    lsl x0, x0, #1
    orr x0, x0, #1
    br x30

    .global fl_float
    // x1=num -> x0
fl_float:
    asr x1, x1, #1
    scvtf d0, x1
    b fl_mkfloat

true:
    adrp x0, fl_true
    add x0, x0, :lo12:fl_true
    ldr x0, [x0]
    br x30

false:
    adrp x0, fl_false
    add x0, x0, :lo12:fl_false
    ldr x0, [x0]
    br x30

    .global fl_sametype
    // x1=val, x2=val -> x0
fl_sametype:
    and x0, x1, #3
    and x8, x2, #3
    cmp x0, x8
    bne false
    tst x0, #3
    bne true
    ldr w0, [x1]
    ldr w8, [x2]
    mov w9, #0x80000000    // FL_TUPLE_BIT
    tst w0, w9
    beq l19
    tst w8, w9
    bne false
    b true
l19:
    tst w8, w9
    bne false
    mov w9, #0xff000000     // FL_TAG_MASK
    and w0, w0, w9
    and w8, w8, w9
    cmp w0, w8
    bne false
    b true

    .global fl_equal
    // x1=val, x2=val -> x0
    // must preserve P0-P3
fl_equal:
    cmp x1, x2
    beq true
    tst x1, #1    
    beq l22
    tst x2, #1
    beq l23
    b false
l22:
    tst x2, #2
    bne false
    ldr w8, [x1]
    mov w9, #0xff000000     // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x02000000     // FL_FLOAT_TAG
    cmp w8, w9
    bne false
l23:
    tst x2, #2
    bne false
    tst x2, #1
    bne l25
    ldr w8, [x2]
    mov w9, #0xff000000     // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x02000000     // FL_FLOAT_TAG
    cmp w8, w9
    bne false
l25:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    fcmp d0, d1
    bne false
    b true
    
    .global fl_less
    // x1=val, x2=val -> x0
    // must preserve P0-P3
fl_less:
    tst x1, #1
    beq l26
    tst x2, #1
    beq l26
    cmp x1, x2
    bge false
    b true
l26:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    fcmp d0, d1
    bmi true
    b false

    .global fl_greater
    // x1=val, x2=val -> x0
    // must preserve P0-P3
fl_greater:
    tst x1, #1
    beq l26b
    tst x2, #1
    beq l26b
    cmp x1, x2
    ble false
    b true
l26b:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    fcmp d1, d0
    bmi true
    b false

first:
    mov x0, x1
    br x30

second:
    mov x0, x2
    br x30

    .global fl_max
    // x1=val, x2=val -> x0
    // must preserve P0-P3
fl_max:
    tst x1, #1
    beq l26a
    tst x2, #1
    beq l26a
    cmp x1, x2
    ble second
    b first
l26a:
    stp x1, x2, [x28, #-16]!
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    ldp x1, x2, [x28], #16
    fcmp d0, d1
    bmi second
    b first

    .global fl_min
    // x1=val, x2=val -> x0
    // must preserve P0-P3
fl_min:
    tst x1, #1
    beq l26i
    tst x2, #1
    beq l26i
    cmp x1, x2
    ble first
    b second
l26i:
    stp x1, x2, [x28, #-16]!
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    ldp x1, x2, [x28], #16
    fcmp d0, d1
    bmi first
    b second

    .global fl_sign
    // x1=val -> x0
fl_sign:
    tst x1, #1
    bne lsign1
    str x30, [x28, #-8]!
    bl float_arg_1
    ldr x30, [x28], #8
    adrp x0, zero
    add x0, x0, :lo12:zero
    ldr d1, [x0]
    fcmp d0, d1
    beq lsign3
    ble lsign2
lsign4:
    mov x0, #3
    br x30
lsign2:
    mov x0, #-1
    br x30
lsign3:
    mov x0, #1
    br x30
lsign1:
    cmp x0, #1
    bmi lsign2
    bne lsign4
    mov x0, x1
    br x30
zero: .double 0

    .global fl_abs
    // x1=val -> x0
fl_abs:
    tst x1, #1
    bne labs1
    str x30, [x28, #-8]!
    bl float_arg_1
    ldr x30, [x28], #8
    fabs d0, d0
    b fl_mkfloat
labs1:
    cmp x1, #0
    bmi labs2
    mov x0, x1
    br x30
labs2:
    neg x0, x1
    add x0, x0, #2
    br x30

    .global fl_rnd
    // x1=val -> x0
fl_rnd:
    mov x0, x1
    b fl_random_int

    .global fl_fip
    // x1=val -> x0
fl_fip:
    tst x1, #1
    beq lfip1
    mov x0, x1
    br x30
lfip1:
    ldr x0, [x1, #8]
    lsr x0, x0, #52
    and x0, x0, #0x7ff
    sub x0, x0, #0x3ff
    cmp x0, #52
    blt lfip2
    mov x0, x1      // nan
    br x30
lfip2:
    cmp x0, #0
    bge lfip4
    mov x0, #1
    br x30
lfip4:
    mov x8, #-1
    lsr x8, x8, #12
    lsr x8, x8, x0
    ldr x0, [x1, #8]
    and x8, x0, x8
    cmp x8, #0
    bne lfip3
    mov x0, x1
    br x30
lfip3:
    mvn x8, x8
    and x0, x8, x0
    fmov d0, x0
    b fl_mkfloat

    .global fl_ffp
    // x1=val -> x0
fl_ffp:
    tst x1, #1
    beq lffp1
    mov x0, #1
    br x30
lffp1:
    ldr x0, [x1, #8]
    lsr x0, x0, #52
    and x0, x0, #0x7ff
    sub x0, x0, #0x3ff
    cmp x0, #52
    blt lffp2
    cmp x0, #0x400
    bne lffp3
    ldr x0, [x1, #8]
    lsl x0, x0, #12
    cmp x0, #0
    beq lffp3
    mov x0, x1      // nan
    br x30
lffp3:
    mov x0, #1
    br x30
lffp2:
    cmp x0, #0
    bge lffp4
    mov x0, x1
    br x30
lffp4:
    mov x8, #-1
    lsr x8, x8, #12
    lsr x8, x8, x0
    ldr x0, [x1, #8]
    and x8, x8, x0
    cmp x8, #0
    beq lffp3
    mvn x8, x8
    and x0, x0, x8
    fmov d1, x0
    ldr d0, [x1, #8]
    fsub d0, d0, d1
    b fl_mkfloat

.global fl_below
    // x1=val, x2=val -> x0
fl_below:
    sub x28, x28, #40
    str x1, [x28, #8]
    str x2, [x28, #16]
    str x3, [x28, #24]
    str x30, [x28, #32]
    mov x0, x19
    mov x3, x28
    bl fl_ordering
    ldr x30, [x28, #32]
    ldr x3, [x28, #24]
    ldr x2, [x28, #16]
    ldr x1, [x28, #8]
    ldr x8, [x28]
    add x28, x28, #40
    tst x8, #1
    beq lbel1
    adrp x0, fl_nil
    add x0, x0, :lo12:fl_nil
    ldr x0, [x0]
    br x30
lbel1:
    tst x0, #2     // -1
    bne true
    b false

    .global fl_above
    // x1=val, x2=val -> x0
fl_above:
    sub x28, x28, #40
    str x1, [x28, #8]
    str x2, [x28, #16]
    str x3, [x28, #24]
    str x30, [x28, #32]
    mov x0, x19
    mov x3, x28
    bl fl_ordering
    ldr x30, [x28, #32]
    ldr x3, [x28, #24]
    ldr x2, [x28, #16]
    ldr x1, [x28, #8]
    ldr x8, [x28]
    add x28, x28, #40
    tst x8, #1
    beq labo1
    adrp x0, fl_nil
    add x0, x0, :lo12:fl_nil
    ldr x0, [x0]
    br x30
labo1:
    cmp x0, #1
    bne false
    b true

    .global fl_deref
    // x0=val -> x0
    // must preserve P0-P3
fl_deref:
    tst x0, #3
    beq l27
l28:
    br x30
l27:
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x04000000 // FL_VAR_TAG
    cmp w8, w9
    bne l28
    ldr x8, [x0, #8]
    cmp x0, x8
    beq l28
    mov x0, x8
    b fl_deref

    .global fl_unref
    // x8=val
    // must preserve A + P0-P3
fl_unref:
    tst x8, #3
    beq l29
    br x30
l29:
    mov w10, #0x00ffffff      // FL_COUNT_MASK
    ldr w11, [x8]
    and w11, w11, w10
    cmp w11, w10
    bne l30
    br x30
l30:
    ldr w11, [x8]
    sub w11, w11, #1
    tst w11, w10
    beq l30a
    str w11, [x8]
    br x30
l30a:
    sub x28, x28, #40
    stp x1, x2, [x28]
    stp x3, x30, [x28, #16]
    str x0, [x28, #32]
    mov x0, x19
    mov x1, x8
    bl fl_release
    ldr x0, [x28, #32]
    ldp x3, x30, [x28, #16]
    ldp x1, x2, [x28]
    add x28, x28, #40
    br x30

setcarry:
    mov x8, #1 << 29
    msr nzcv, x8
    br x30

clrcarry:
    mrs x8, nzcv
    and x8, x8, #~(1 << 29)
    msr nzcv, x8
    br x30

    .global fl_force
    // x0=val -> x0, CF is 1 when suspending
fl_force:
    tst x0, #3
    bne clrcarry
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x07000000 // FL_REF_TAG
    cmp w8, w9
    beq l33
    mov w9, #0x04000000 // FL_VAR_TAG
    cmp w8, w9
    bne clrcarry
    ldr x8, [x0, #8]
    cmp x0, x8
    beq l33
    mov x0, x8
    b fl_force
l33:
    ldr x8, [x19, #8]   // tcb->sstackp
    str x0, [x8]
    ldr x0, [x20]      // goal->addr
    str x0, [x8, #8]
    add x8, x8, #16
    str x8, [x19, #8]   // tcb->sstackp
    b setcarry

    .global fl_match
    // x1=val1, x2=val2 -> x0
    // assumes args are forced (allows initial quick eq check)
fl_match:
    cmp x1, x2
    beq true
    mov x0, x19
    sub x28, x28, #32
    stp x1, x30, [x28]
    str x2, [x28, #16]
    bl fl_match_rec
    ldr x2, [x28, #16]
    ldp x1, x30, [x28]
    add x28, x28, #32
    br x30

    .global fl_islist_t
    // x0=val, set CF if not a list
fl_islist_t:
    tst x0, #3
    bne setcarry
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x01000000 // FL_LIST_TAG
    cmp w8, w9
    beq clrcarry
    b setcarry

    .global fl_istuple_t
    // x0=val, set CF if not a tuple
fl_istuple_t:
    tst x0, #3
    bne setcarry
    ldr w8, [x0]
    mov w9, #0x80000000 // FL_TUPLE_BIT
    tst w8, w9
    bne clrcarry
    b setcarry

    // x1=addr -> Carry (fail)
trynext:
    str x1, [x20]      // goal->addr
    b setcarry

    .global fl_isval
    // x0=val, x2=atom, x1=addr
fl_isval:
    cmp x0, x2
    bne trynext
    b clrcarry

    .global fl_isfloat
    // x0=val, d0=float, x1=addr 
fl_isfloat:
    tst x0, #3
    bne trynext
    ldr d1, [x0, #8]
    fcmp d0, d1
    bne trynext
    b clrcarry

    .global fl_islist
    // x0=val, x1=addr -> x0(car)
    // push cdr on stack
fl_islist:
    str x30, [x28, #-8]!
    bl fl_islist_t
    ldr x30, [x28], #8
    bcs trynext
    ldr x8, [x0, #16]
    str x8, [x28, #-8]!
    ldr x0, [x0, #8]
    b clrcarry

    .global fl_istuple
    // x0=val, x1=addr, x2=arity -> x0(head)
    // push arglist on stack
fl_istuple:
    tst x0, #3
    bne trynext
    ldr w8, [x0]
    mov w9, #0x80000000 // FL_TUPLE_BIT
    tst w8, w9
    beq trynext
    lsr w8, w8, #24
    and w8, w8, #0x7f
    cmp w8, w2
    bne trynext
    ldr x3, [x0, #16]
    str x3, [x28, #-8]!
    ldr x0, [x0, #8]
    b clrcarry

    // x8=car, x0=cdr -> x0
    // must preserve parameter registers
cons:    
    ldr x9, [x19, #6 * 8]   // tcb->freelist
    cmp x9, #0
    bne l34
    adrp x1, fl_nil
    add x1, x1, :lo12:fl_nil
    ldr x1, [x11]
    mov x0, x19
    mov x2, #10             // FL_NO_CELLS
    b fl_rt_error
l34:
    str x30, [x28, #-8]!
    ldr x10, [x9, #16]
    str x10, [x19, #6 * 8]  // tcb->freelist
    bl fl_addref
    str x0, [x9, #16]
    mov x0, x8
    bl fl_addref
    str x0, [x9, #8]
    ldr x10, [x19, #11 * 8]  // tcb->used
    add x10, x10, #1
    str x10, [x19, #11 * 8]    // tcb->used
    mov x0, x9
    ldr x30, [x28], #8
    br x30

    .global fl_mkfloat
    // d0=float64 -> x0
    // must preserve parameter registers
fl_mkfloat:
    str x30, [x28, #-8]!
    mov x8, #1
    mov x0, #1
    bl cons
    str d0, [x0, #8]
    mov x8, #0x02000000      // FL_FLOAT_TAG
    str w8, [x0]
    ldr x30, [x28], #8
    br x30

    .global fl_mklist
    // x0=cdr, <stack>=car -> x0
fl_mklist:
    ldr x8, [x28]
    str x30, [x28]
    bl cons
    mov x8, #0x01000000      // FL_LIST_TAG
    str w8, [x0]
    ldr x30, [x28], #8
    br x30

    .global fl_mktuple
    // x0=args, x8=arity, x9=atom -> x0
fl_mktuple:
    stp x8, x30, [x28, #-16]!
    mov x8, x9
    bl cons
    ldp x8, x30, [x28], #16
    // fall through ...

    .global fl_mktuple0
    // x0=args, x8=arity -> x0
fl_mktuple0:
    lsl x8, x8, #24
    mov w9, #0x80000000 // FL_TUPLE_BIT
    orr w8, w8, w9
    str w8, [x0]
    br x30

    .global fl_mkvar
    // -> x0
fl_mkvar:
    str x30, [x28, #-8]!
    mov x8, #1
    adrp x0, fl_nil
    add x0, x0, :lo12:fl_nil
    ldr x0, [x0]
    bl cons
    str x0, [x0, #8]
    mov x8, #0x04000000   // FL_VAR_TAG
    str w8, [x0]
    ldr x30, [x28], #8
    br x30

    // x21=goal
enqueue:
    ldr x0, [x19, #13 * 8]     // tcb->qend
    str x21, [x0]
    add x0, x0, #8
    ldr x3, [x19, #12 * 8]     // tcb->qstart
    cmp x0, x3
    bne l40
    adrp x1, fl_nil
    add x1, x1, :lo12:fl_nil
    ldr x1, [x1]
    mov x2, #7                  // FL_QUEUE_FULL
    mov x0, x19
    b fl_rt_error
l40:
    ldr x8, [x19, #15 * 8]     // tcb->max_goals
    lsl x8, x8, #3
    ldr x9, [x19, #14 * 8]     // tcb->queue
    add x8, x8, x9
    cmp x0, x8
    blo l41
    ldr x0, [x19, #14 * 8]     // tcb->queue
l41:
    str x0, [x19, #13 * 8]     // tcb->qend
    ldr x8, [x19, #10 * 8]     // tcb->active
    add x8, x8, #1
    str x8, [x19, #10 * 8]    // tcb->active
    br x30

    .global fl_opengoal
    // -> x21
fl_opengoal:
    ldr x21, [x19, #16 * 8]     // tcb->freegoals
    cmp x21, #0
    bne lg1
    mov x2, #26                  // FL_TOO_MANY_GOALS
    mov x0, x19
    b fl_rt_error
lg1:
    ldr x8, [x21]               // goal->addr
    ldr x0, [x20, #6 * 8]       // goal->task
    str x30, [x28, #-8]!
    bl fl_addref
    ldr x30, [x28], #8
    str x0, [x21, #6 * 8]       // goal->task
    str x8, [x19, #16 * 8]      // tcb->freegoals
    ldr x8, [x19, #21 * 8]    // tcb->goals
    add x8, x8, #1
    str x8, [x19, #21 * 8]    // tcb->goals
    b enqueue

    .global fl_dropgoal
fl_dropgoal:
    ldr x8, [x19]               // tcb->goal
    ldr x9, [x19, #16 * 8]     // tcb->freegoals
    str x9, [x8]               // goal->addr
    str x8, [x19, #16 * 8]      // tcb->freegoals
    ldr x9, [x19, #21 * 8]    // tcb->goals
    sub x9, x9, #1
    str x9, [x19, #21 * 8]    // tcb->goals
    adrp x1, fl_nil 
    add x1, x1, :lo12:fl_nil
    ldr x1, [x1]
    ldr x0, [x8, #6 * 8]      // goal->task
    str x1, [x8, #6 * 8]      // goal->task
    mov x8, x0
    b fl_unref

    .global fl_addref
    // x0=val
    // may not change any registers
fl_addref:
    tst x0, #3
    beq l42
    br x30
l42:
    stp x8, x9, [x28, #-16]!
    mov w9, #0x00ffffff     // FL_COUNT_MASK
    ldr w8, [x0]
    and w8, w8, w9
    cmp w8, w9
    beq l42a
    ldr w8, [x0]
    add w8, w8, #1
    str w8, [x0]
l42a:
    ldp x8, x9, [x28], #16
    br x30

    .global fl_puttailarg
    // x0=list
fl_puttailarg:
    ldr x8, [x0, #16]
    tst x8, #3
    beq l43
    ldr x0, [x0, #8]
l43:
    str x30, [x28, #-8]!
    bl fl_addref
    str x0, [x21, #4 * 8]   // goal->args[ 3 ]
    ldr x30, [x28], #8
    br x30

    .global fl_resolve
    // x4=cache, x1=arity, x2=module, x3=name
fl_resolve:
    ldr x8, [x4]
    cmp x8, #0
    beq l44
    str x8, [x21]        // goal->addr
    br x30
l44:
    mov x0, x19
    mov x5, x21
    b fl_resolve_pdef

    .global fl_replace
    // x0=val, x2=argaddr
fl_replace:
    ldr x8, [x2]
    cmp x0, x8
    bne l45
    br x30
l45:
    ldr x8, [x2]
    str x30, [x28, #-8]!
    bl fl_addref
    ldr x30, [x28], #8
    str x0, [x2]
    b fl_unref
    
    .global fl_tailgoal
    // x2=addr -> x21(N)
fl_tailgoal:
    ldr x3, [x19, #16] // tcb->tailcalls
    ldr x1, [x19, #24] // tcb->timeslice
    cmp x3, x1
    blt l_tg_1
    stp x2, x30, [x28, #-16]!
    bl fl_opengoal
    ldp x2, x30, [x28], #16
    str x2, [x21]      // goal->addr
    br x30
l_tg_1:
    mov x21, x20
    str x2, [x21]      // goal->addr
    br x30

    .global fl_tail_jump
fl_tail_jump:
    ldr x3, [x19, #16] // tcb->tailcalls
    ldr x1, [x19, #24] // tcb->timeslice
    cmp x3, x1
    blt l46
    mov x0, #0
    str x0, [x19, #16] // tcb->tailcalls
    br x30
l46:
    add x3, x3, #1
    str x3, [x19, #16] // tcb->tailcalls
    // fall through ...

    .global fl_invoke_jump
fl_invoke_jump:
    mov x28, x29
    mov x20, x21
    str x21, [x19]      // tcb->goal
    ldr x0, [x21]      // goal->addr
    br x0

    .global fl_test_integer
    // x1=val -> x0 (bool)
fl_test_integer:
    tst x1, #1
    bne true
    b false

    .global fl_test_number
    // x1=val -> x0 (bool)
fl_test_number:
    tst x1, #1
    bne true
    mov w8, #0x02000000      // FL_FLOAT_TAG
    b fl_test_tag

    .global fl_test_atom
    // x1=val -> x0 (bool)
fl_test_atom:
    mov x0, x1
    and x0, x0, #3
    cmp x0, #2
    bne false
    b true

    .global fl_test_tag
    // x1=val, w8=tag -> x0 (bool)
fl_test_tag:
    tst x1, #3
    bne false
    ldr w0, [x1]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w0, w0, w9
    cmp w0, w8
    bne false
    b true

    .global fl_test_var
    // x1=val -> x0 (bool)
fl_test_var:
    tst x1, #3
    bne false
    ldr w0, [x1]
    mov w8, #0xff000000 // FL_TAG_MASK
    and w0, w0, w8
    mov w8, #0x07000000 // FL_REF_TAG
    cmp w0, w8
    beq true
    mov w8, #0x04000000 // FL_VAR_TAG
    cmp w0, w8
    bne false
    ldr x0, [x1, #8]
    cmp x0, x1
    bne false
    b true

    .global fl_test_remote
    // x1=val -> x0 (bool)
fl_test_remote:
    tst x1, #3
    bne false
    ldr w0, [x1]
    mov w8, #0xff000000 // FL_TAG_MASK
    and w0, w0, w8
    mov w8, #0x07000000 // FL_REF_TAG
    cmp w0, w8
    beq true
    mov w8, #0x05000000 // FL_PORT_TAG
    cmp w0, w8
    bne false
    ldr x0, [x1, #8]
    tst x0, #1
    bne true
    b false

    .global fl_test_tuple
    // x1=val -> x0 (bool)
fl_test_tuple:
    tst x1, #3
    bne false
    ldr w0, [x1]
    mov w8, #0x80000000 // FL_TUPLE_BIT
    tst w0, w8
    beq false
    b true

    .global fl_enter
    // x0=tcb
fl_enter:
    mov x19, x0
    ldr x28, [x19, #5 * 8]  // tcb->stackbase
    ldr x0, [x19, #22*8]    // tcb->sp0
    mov sp, x0
    mov x29, x28
    ldr x20, [x19]      // tcb->goal
    ldr x0, [x20]      // goal->addr
    br x0

    .global fl_atomic
    // x1=val -> x0 (bool)
fl_atomic:
    tst x1, #3         // FL_BITS_MASK
    bne true
    ldr w0, [x1]
    mov w8, #0xff000000     // FL_TAG_MASK
    and w0, w0, w8
    mov w8, #0x02000000     // FL_FLOAT_TAG
    cmp w0, w8
    bne false
    b true

    .global fl_entry
    // x1=arity
fl_entry:
    ldr x0, [x19, #20 * 8]     // tcb->logging
    tst x0, #1
    bne l50
    br x30
l50:
    mov x0, x19
    b fl_log_entry

    .global fl_assign0
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
fl_assign0:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl fl_assign
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    br x30

    .global fl_d_assign0
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
fl_d_assign0:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl fl_d_assign
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    br x30

    .global fl_unify
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
fl_unify:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl fl_unify_rec
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    br x30

    .global fl_d_unify
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
fl_d_unify:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl fl_d_unify_rec
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    br x30

    .global fl_unify_safe
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
fl_unify_safe:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl fl_unify_safe1
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    br x30

    .global fl_unify_checked
    // x1=val, x2=val
    // must preserve P0-P3
fl_unify_checked:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl fl_unify_rec_checked
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    br x30

    .global fl_d_unify_checked
    // x1=val, x2=val
    // must preserve P0-P3
fl_d_unify_checked:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl fl_d_unify_rec_checked
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    br x30
